@@ -1,20 +1,19 @@
 /*
  * Copyright 2014-2016 Milian Wolff <mail@milianw.de>
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Library General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 /**
@@ -86,6 +85,34 @@ ostream& operator<<(ostream& out, const formatBytes data)
     return out << fixed << setprecision(2) << bytes << *unit;
 }
 
+enum CostType
+{
+    Allocations,
+    Temporary,
+    Allocated,
+    Leaked,
+    Peak
+};
+
+std::istream& operator>>(std::istream& in, CostType& type)
+{
+    std::string token;
+    in >> token;
+    if (token == "allocations")
+        type = Allocations;
+    else if (token == "temporary")
+        type = Temporary;
+    else if (token == "allocated")
+        type = Allocated;
+    else if (token == "leaked")
+        type = Leaked;
+    else if (token == "peak")
+        type = Peak;
+    else
+        in.setstate(std::ios_base::failbit);
+    return in;
+}
+
 struct Printer final : public AccumulatedTraceData
 {
     void finalize()
@@ -153,12 +180,21 @@ struct Printer final : public AccumulatedTraceData
                                         auto node = findTrace(allocation.traceIndex);
                                         while (node.ipIndex) {
                                             const auto& ip = findIp(node.ipIndex);
-                                            if (isStopIndex(ip.functionIndex)) {
+                                            if (isStopIndex(ip.frame.functionIndex)) {
                                                 break;
                                             }
-                                            if (stringify(ip.functionIndex).find(filterBtFunction) != string::npos) {
+                                            auto matchFunction = [this](const Frame& frame) {
+                                                return stringify(frame.functionIndex).find(filterBtFunction)
+                                                    != string::npos;
+                                            };
+                                            if (matchFunction(ip.frame)) {
                                                 return false;
                                             }
+                                            for (const auto& inlined : ip.inlined) {
+                                                if (matchFunction(inlined)) {
+                                                    return false;
+                                                }
+                                            }
                                             node = findTrace(node.parentIndex);
                                         };
                                         return true;
@@ -182,28 +218,36 @@ struct Printer final : public AccumulatedTraceData
     {
         printIndent(out, indent);
 
-        if (ip.functionIndex) {
-            out << prettyFunction(stringify(ip.functionIndex));
+        if (ip.frame.functionIndex) {
+            out << prettyFunction(stringify(ip.frame.functionIndex));
         } else {
             out << "0x" << hex << ip.instructionPointer << dec;
         }
 
         if (flameGraph) {
             // only print the file name but nothing else
-            if (ip.fileIndex) {
-                const auto& file = stringify(ip.fileIndex);
+            auto printFile = [this, &out](FileIndex fileIndex) {
+                const auto& file = stringify(fileIndex);
                 auto idx = file.find_last_of('/') + 1;
                 out << " (" << file.substr(idx) << ")";
+            };
+            if (ip.frame.fileIndex) {
+                printFile(ip.frame.fileIndex);
             }
             out << ';';
+            for (const auto& inlined : ip.inlined) {
+                out << prettyFunction(stringify(inlined.functionIndex));
+                printFile(inlined.fileIndex);
+                out << ';';
+            }
             return;
         }
 
         out << '\n';
         printIndent(out, indent + 1);
 
-        if (ip.fileIndex) {
-            out << "at " << stringify(ip.fileIndex) << ':' << ip.line << '\n';
+        if (ip.frame.fileIndex) {
+            out << "at " << stringify(ip.frame.fileIndex) << ':' << ip.frame.line << '\n';
             printIndent(out, indent + 1);
         }
 
@@ -213,6 +257,13 @@ struct Printer final : public AccumulatedTraceData
             out << "in ??";
         }
         out << '\n';
+
+        for (const auto& inlined : ip.inlined) {
+            printIndent(out, indent);
+            out << prettyFunction(stringify(inlined.functionIndex)) << '\n';
+            printIndent(out, indent + 1);
+            out << "at " << stringify(inlined.fileIndex) << ':' << inlined.line << '\n';
+        }
     }
 
     void printBacktrace(const TraceIndex traceIndex, ostream& out, const size_t indent = 0,
@@ -227,6 +278,8 @@ struct Printer final : public AccumulatedTraceData
 
     void printBacktrace(TraceNode node, ostream& out, const size_t indent = 0, bool skipFirst = false) const
     {
+        unordered_set<uint32_t> recursionGuard;
+        recursionGuard.insert(node.ipIndex.index);
         while (node.ipIndex) {
             const auto& ip = findIp(node.ipIndex);
             if (!skipFirst) {
@@ -234,7 +287,12 @@ struct Printer final : public AccumulatedTraceData
             }
             skipFirst = false;
 
-            if (isStopIndex(ip.functionIndex)) {
+            if (isStopIndex(ip.frame.functionIndex)) {
+                break;
+            }
+
+            if (!recursionGuard.insert(node.parentIndex.index).second) {
+                cerr << "Trace recursion detected - corrupt data file?" << endl;
                 break;
             }
 
@@ -255,7 +313,7 @@ struct Printer final : public AccumulatedTraceData
 
         const auto& ip = findIp(node.ipIndex);
 
-        if (!isStopIndex(ip.functionIndex)) {
+        if (!isStopIndex(ip.frame.functionIndex)) {
             printFlamegraph(findTrace(node.parentIndex), out);
         }
         printIp(ip, out, 0, true);
@@ -290,6 +348,9 @@ struct Printer final : public AccumulatedTraceData
             int64_t handled = 0;
             for (size_t j = 0; j < min(subPeakLimit, allocation.traces.size()); ++j) {
                 const auto& trace = allocation.traces[j];
+                if (!(trace.*member)) {
+                    break;
+                }
                 sublabel(trace);
                 handled += trace.*member;
                 printBacktrace(trace.traceIndex, cout, 2, true);
@@ -371,7 +432,7 @@ struct Printer final : public AccumulatedTraceData
         const auto ip = findIp(location);
 
         // skip anything below main
-        const bool shouldStop = isStopIndex(ip.functionIndex);
+        const bool shouldStop = isStopIndex(ip.frame.functionIndex);
         if (!shouldStop) {
             for (auto& merged : mergedAllocations) {
                 if (merged.leaked < 0) {
@@ -395,6 +456,7 @@ struct Printer final : public AccumulatedTraceData
             }
         }
 
+        // TODO: write inlined frames out to massif files
         printIndent(massifOut, depth, " ");
         massifOut << 'n' << (numAllocs + (skipped ? 1 : 0)) << ": " << heapSize;
         if (!depth) {
@@ -402,15 +464,15 @@ struct Printer final : public AccumulatedTraceData
                          "--alloc-fns, etc.\n";
         } else {
             massifOut << " 0x" << hex << ip.instructionPointer << dec << ": ";
-            if (ip.functionIndex) {
-                massifOut << stringify(ip.functionIndex);
+            if (ip.frame.functionIndex) {
+                massifOut << stringify(ip.frame.functionIndex);
             } else {
                 massifOut << "???";
             }
 
             massifOut << " (";
-            if (ip.fileIndex) {
-                massifOut << stringify(ip.fileIndex) << ':' << ip.line;
+            if (ip.frame.fileIndex) {
+                massifOut << stringify(ip.frame.fileIndex) << ':' << ip.frame.line;
             } else if (ip.moduleIndex) {
                 massifOut << stringify(ip.moduleIndex);
             } else {
@@ -442,7 +504,7 @@ struct Printer final : public AccumulatedTraceData
         }
     }
 
-    void handleAllocation(const AllocationInfo& info, const AllocationIndex /*index*/) override
+    void handleAllocation(const AllocationInfo& info, const AllocationInfoIndex /*index*/) override
     {
         if (printHistogram) {
             ++sizeHistogram[info.size];
@@ -492,52 +554,61 @@ struct Printer final : public AccumulatedTraceData
 int main(int argc, char** argv)
 {
     po::options_description desc("Options", 120, 60);
-    desc.add_options()("file,f", po::value<string>(), "The heaptrack data file to print.")(
-        "diff,d", po::value<string>()->default_value({}), "Find the differences to this file.")(
-        "shorten-templates,t", po::value<bool>()->default_value(true)->implicit_value(true),
-        "Shorten template identifiers.")("merge-backtraces,m",
-                                         po::value<bool>()->default_value(true)->implicit_value(true),
-                                         "Merge backtraces.\nNOTE: the merged peak consumption is not correct.")(
-        "print-peaks,p", po::value<bool>()->default_value(true)->implicit_value(true),
-        "Print backtraces to top allocators, sorted by peak consumption.")(
-        "print-allocators,a", po::value<bool>()->default_value(true)->implicit_value(true),
-        "Print backtraces to top allocators, sorted by number of calls to "
-        "allocation functions.")("print-temporary,T", po::value<bool>()->default_value(true)->implicit_value(true),
-                                 "Print backtraces to top allocators, sorted by number of temporary "
-                                 "allocations.")("print-leaks,l",
-                                                 po::value<bool>()->default_value(false)->implicit_value(true),
-                                                 "Print backtraces to leaked memory allocations.")(
-        "print-overall-allocated,o", po::value<bool>()->default_value(false)->implicit_value(true),
-        "Print top overall allocators, ignoring memory frees.")(
-        "peak-limit,n", po::value<size_t>()->default_value(10)->implicit_value(10),
-        "Limit the number of reported peaks.")("sub-peak-limit,s",
-                                               po::value<size_t>()->default_value(5)->implicit_value(5),
-                                               "Limit the number of reported backtraces of merged peak locations.")(
-        "print-histogram,H", po::value<string>()->default_value(string()),
-        "Path to output file where an allocation size histogram will be written "
-        "to.")("print-flamegraph,F", po::value<string>()->default_value(string()),
-               "Path to output file where a flame-graph compatible stack file will be "
-               "written to.\n"
-               "To visualize the resulting file, use flamegraph.pl from "
-               "https://github.com/brendangregg/FlameGraph:\n"
-               "  heaptrack_print heaptrack.someapp.PID.gz -F stacks.txt\n"
-               "  # optionally pass --reverse to flamegraph.pl\n"
-               "  flamegraph.pl --title \"heaptrack: allocations\" --colors mem \\\n"
-               "    --countname allocations < stacks.txt > heaptrack.someapp.PID.svg\n"
-               "  [firefox|chromium] heaptrack.someapp.PID.svg\n")(
-        "print-massif,M", po::value<string>()->default_value(string()),
-        "Path to output file where a massif compatible data file will be written "
-        "to.")("massif-threshold", po::value<double>()->default_value(1.),
-               "Percentage of current memory usage, below which allocations are "
-               "aggregated into a 'below threshold' entry.\n"
-               "This is only used in the massif output file so far.\n")(
-        "massif-detailed-freq", po::value<size_t>()->default_value(2),
-        "Frequency of detailed snapshots in the massif output file. Increase "
-        "this to reduce the file size.\n"
-        "You can set the value to zero to disable detailed snapshots.\n")(
-        "filter-bt-function", po::value<string>()->default_value(string()),
-        "Only print allocations where the backtrace contains the given "
-        "function.")("help,h", "Show this help message.")("version,v", "Displays version information.");
+    // clang-format off
+    desc.add_options()
+        ("file,f", po::value<string>(),
+            "The heaptrack data file to print.")
+        ("diff,d", po::value<string>()->default_value({}),
+            "Find the differences to this file.")
+        ("shorten-templates,t", po::value<bool>()->default_value(true)->implicit_value(true),
+            "Shorten template identifiers.")
+        ("merge-backtraces,m", po::value<bool>()->default_value(true)->implicit_value(true),
+            "Merge backtraces.\nNOTE: the merged peak consumption is not correct.")
+        ("print-peaks,p", po::value<bool>()->default_value(true)->implicit_value(true),
+            "Print backtraces to top allocators, sorted by peak consumption.")
+        ("print-allocators,a", po::value<bool>()->default_value(true)->implicit_value(true),
+            "Print backtraces to top allocators, sorted by number of calls to allocation functions.")
+        ("print-temporary,T", po::value<bool>()->default_value(true)->implicit_value(true),
+            "Print backtraces to top allocators, sorted by number of temporary allocations.")
+        ("print-leaks,l", po::value<bool>()->default_value(false)->implicit_value(true),
+            "Print backtraces to leaked memory allocations.")
+        ("print-overall-allocated,o", po::value<bool>()->default_value(false)->implicit_value(true),
+            "Print top overall allocators, ignoring memory frees.")
+        ("peak-limit,n", po::value<size_t>()->default_value(10)->implicit_value(10),
+            "Limit the number of reported peaks.")
+        ("sub-peak-limit,s", po::value<size_t>()->default_value(5)->implicit_value(5),
+            "Limit the number of reported backtraces of merged peak locations.")
+        ("print-histogram,H", po::value<string>()->default_value(string()),
+            "Path to output file where an allocation size histogram will be written to.")
+        ("flamegraph-cost-type", po::value<CostType>()->default_value(Allocations),
+            "The cost type to use when generating a flamegraph. Possible options are:\n"
+            "  - allocations: number of allocations\n"
+            "  - temporary: number of temporary allocations\n"
+            "  - allocated: bytes allocated, ignoring deallocations\n"
+            "  - leaked: bytes not deallocated at the end\n"
+            "  - peak: bytes consumed at highest total memory consumption")
+        ("print-flamegraph,F", po::value<string>()->default_value(string()),
+            "Path to output file where a flame-graph compatible stack file will be written to.\n"
+            "To visualize the resulting file, use flamegraph.pl from "
+            "https://github.com/brendangregg/FlameGraph:\n"
+            "  heaptrack_print heaptrack.someapp.PID.gz -F stacks.txt\n"
+            "  # optionally pass --reverse to flamegraph.pl\n"
+            "  flamegraph.pl --title \"heaptrack: allocations\" --colors mem \\\n"
+            "    --countname allocations < stacks.txt > heaptrack.someapp.PID.svg\n"
+            "  [firefox|chromium] heaptrack.someapp.PID.svg\n")
+        ("print-massif,M", po::value<string>()->default_value(string()),
+            "Path to output file where a massif compatible data file will be written to.")
+        ("massif-threshold", po::value<double>()->default_value(1.),
+            "Percentage of current memory usage, below which allocations are aggregated into a 'below threshold' entry.\n"
+            "This is only used in the massif output file so far.\n")
+        ("massif-detailed-freq", po::value<size_t>()->default_value(2),
+            "Frequency of detailed snapshots in the massif output file. Increase  this to reduce the file size.\n"
+            "You can set the value to zero to disable detailed snapshots.\n")
+        ("filter-bt-function", po::value<string>()->default_value(string()),
+            "Only print allocations where the backtrace contains the given function.")
+        ("help,h", "Show this help message.")
+        ("version,v", "Displays version information.");
+    // clang-format on
     po::positional_options_description p;
     p.add("file", -1);
 
@@ -584,6 +655,7 @@ int main(int argc, char** argv)
     const string printHistogram = vm["print-histogram"].as<string>();
     data.printHistogram = !printHistogram.empty();
     const string printFlamegraph = vm["print-flamegraph"].as<string>();
+    const auto flamegraphCostType = vm["flamegraph-cost-type"].as<CostType>();
     const string printMassif = vm["print-massif"].as<string>();
     if (!printMassif.empty()) {
         data.massifOut.open(printMassif, ios_base::out);
@@ -650,18 +722,7 @@ int main(int argc, char** argv)
     }
 
     if (printPeaks) {
-        /// FIXME: find a way to merge this without breaking temporal dependency.
-        /// I.e. a given function could be called N times from different places
-        /// and allocate M bytes each, but free it thereafter.
-        /// Then the below would give a wrong total peak size of N * M instead
-        /// of just N!
         cout << "PEAK MEMORY CONSUMERS\n";
-        if (data.mergeBacktraces) {
-            cout << "\nWARNING - the data below is not an accurate calcuation of"
-                    " the total peak consumption and can easily be wrong.\n"
-                    " For an accurate overview, disable backtrace merging.\n";
-        }
-
         data.printAllocations(&AllocationData::peak,
                               [](const AllocationData& data) {
                                   cout << formatBytes(data.peak) << " peak memory consumed over " << data.allocations
@@ -671,6 +732,7 @@ int main(int argc, char** argv)
                                   cout << formatBytes(data.peak) << " consumed over " << data.allocations
                                        << " calls from:\n";
                               });
+        cout << endl;
     }
 
     if (printLeaks) {
@@ -694,12 +756,12 @@ int main(int argc, char** argv)
         data.printAllocations(&AllocationData::temporary,
                               [](const AllocationData& data) {
                                   cout << data.temporary << " temporary allocations of " << data.allocations
-                                       << " allocations in total (" << setprecision(2)
+                                       << " allocations in total (" << fixed << setprecision(2)
                                        << (float(data.temporary) * 100.f / data.allocations) << "%) from\n";
                               },
                               [](const AllocationData& data) {
                                   cout << data.temporary << " temporary allocations of " << data.allocations
-                                       << " allocations in total (" << setprecision(2)
+                                       << " allocations in total (" << fixed << setprecision(2)
                                        << (float(data.temporary) * 100.f / data.allocations) << "%) from:\n";
                               });
         cout << endl;
@@ -739,7 +801,25 @@ int main(int argc, char** argv)
                 } else {
                     data.printFlamegraph(data.findTrace(allocation.traceIndex), flamegraph);
                 }
-                flamegraph << ' ' << allocation.allocations << '\n';
+                flamegraph << ' ';
+                switch (flamegraphCostType) {
+                case Allocations:
+                    flamegraph << allocation.allocations;
+                    break;
+                case Allocated:
+                    flamegraph << allocation.allocated;
+                    break;
+                case Temporary:
+                    flamegraph << allocation.temporary;
+                    break;
+                case Peak:
+                    flamegraph << allocation.peak;
+                    break;
+                case Leaked:
+                    flamegraph << allocation.leaked;
+                    break;
+                }
+                flamegraph << '\n';
             }
         }
     }
